assert(LibStub, "Broker_PlayTimer requires LibStub")

local db, sessionTime, startOfNewDay, timerDisplayed
local player, realm = UnitName("player"), GetRealmName()
local showingTooltip = false
local blizzTotalPlayed, blizzLevelPlayed = 0, 0

local WHITE 	= {r=1.0, g=1.0, b=1.0}
local BLACK		= {r=0.0, g=0.0, b=0.0}
local RED		= {r=1.0, g=0.0, b=0.0}
local GREEN		= {r=0.0, g=1.0, b=0.0}
local BLUE		= {r=0.0, g=0.0, b=1.0}
local YELLOW	= {r=1.0, g=1.0, b=0.0}
local CYAN		= {r=0.0, g=1.0, b=1.0}
local LTBLUE	= {r=0.5, g=0.5, b=1.0}
local PURPLE	= {r=1.0, g=0.5, b=1.0}
local ORANGE	= {r=1.0, g=0.5, b=.25}

local ALWAYS			= 1
local IF_BOTH_PLAYED	= 2
local NEVER				= 3

local SESSION_TIMER		= 1
local TODAY_TIMER		= 2
local PLAYER_TIMER		= 3

local SINGULAR			= true
local PLURAL			= false

local SHORT_TIME_FORMAT	= 1
local MED_TIME_FORMAT	= 2
local LONG_TIME_FORMAT	= 3

local timerNames = {
		"Session",
		"Today",
		player,
	}

local WHISPER			= 1
local RAID_WARNING		= 2

local factionColor = { ["Alliance"] = CYAN, ["Horde"] = RED }

local playTimer = LibStub("AceAddon-3.0"):NewAddon("PlayTimer", 'AceTimer-3.0', 'AceEvent-3.0')
local playTimerOptionsMenu
local dataobj = LibStub("LibDataBroker-1.1"):NewDataObject("PlayTimer", {
	type 	= "data source",
	icon 	= [[Interface\Icons\Spell_Holy_BorrowedTime]],
	label 	= "PlayTimer",
	text 	= "0m",
})


function playTimer:OnEnable()
	db = LibStub("AceDB-3.0"):New("Broker_PlayTimerDB", {
		global = {
			todayTimer 			= 0,
			lastTimerUpdate 	= time(),
			realmTable 			= {},
			
			remindersEnabled	= false,
			reminderTimer		= TODAY_TIMER,
			remindMinutes		= 60,
			snoozeEnabled		= false,
			snoozeMinutes		= 30,
			reminderType		= RAID_WARNING,
			reminderText		= "Reminder: You have been playing for %s.",
		},
		profile = {
			timeFormat			= SHORT_TIME_FORMAT,
			showColorText 		= true,
			showTimerName		= true,
			showFactions 		= IF_BOTH_PLAYED,
			defaultTimer 		= TODAY_TIMER,
			firstHourOfNewDay	= 4,
		},
	})
	
	local brokerOptions = LibStub("AceConfigRegistry-3.0"):GetOptionsTable("Broker", "dialog", "LibDataBroker-1.1")
	if (not brokerOptions) then
		brokerOptions = {
			type = "group",
			name = "Broker",
			args = {
			}
		}
		LibStub("AceConfig-3.0"):RegisterOptionsTable("Broker", brokerOptions)
		LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Broker", "Broker")
	end
	
	LibStub("AceConfig-3.0"):RegisterOptionsTable("Broker_PlayTimer", {
		name = "Broker_PlayTimer",
		handler = playTimer,
		type = "group",
		args = {
			playTimerDescription = {
				type = 'description',
				name = "A Data Broker plugin to display the amount of time you've been playing.",
				order = 0,
			},			
			displayOptions = {
				name = "Display",
				type = "header",
				order = 10,
			},
			timeFormat = {
				type = 'select',
				style = 'dropdown',
				values = {						-- "83" == 1 hour, 23 minutes
					[SHORT_TIME_FORMAT]	= formatTime(83, SHORT_TIME_FORMAT),
					[MED_TIME_FORMAT]	= formatTime(83, MED_TIME_FORMAT),
					[LONG_TIME_FORMAT]	= formatTime(83, LONG_TIME_FORMAT),
				},
				name = "Time format:",
				desc = "Times will be displayed in the selected format.",
				get = function(_) return db.profile.timeFormat end,
				set = function(_, value) 
					db.profile.timeFormat = value
					UpdateStatusBarText()
				end,
				order = 20,
			},
			colorText = {
				type = 'toggle',
				name = "Use color",
				desc = "Controls the use of color in the tooltip and on the status bar display.",
				get = function(_) return db.profile.showColorText end,
				set = function(_, value) 
					db.profile.showColorText = value
					UpdateStatusBarText()
				end,
				order = 30,
			},
			defaultTimer = {
				type = 'select',
				style = 'dropdown',
				values = { "Time played this session", "Total played today", "This character today" },
				name = "Default status bar timer:",
				desc = "The selected timer will be displayed in the status bar when the player initially enters the realm.",
				get = function(_) return db.profile.defaultTimer end,
				set = function(_, value) 
					db.profile.defaultTimer = value
					timerDisplayed = value
					UpdateStatusBarText()
				end,
				order = 40,
			},
			timerName = {
				type = 'toggle',
				name = "Show name of timer",
				desc = "Show the timer's name in the status bar display.",
				get = function(_) return db.profile.showTimerName end,
				set = function(_, value) 
					db.profile.showTimerName = value
					UpdateStatusBarText()					
					end,
				order = 50,
			},
			newDayStart = {
				type = 'select',
				style = 'dropdown',
				values = { "midnight", "1:00 AM", "2:00 AM", "3:00 AM", "4:00 AM", "5:00 AM", "6:00 AM" },
				name = "Reset timers at:",
				desc = "Player, realm and faction timers will be reset to zero minutes at the selected time of day.",
				get = function(_) return db.profile.firstHourOfNewDay + 1 end,
				set = function(_, value) db.profile.firstHourOfNewDay = value - 1 end,
				order = 60,
			},
			showFactions = {
				type = 'select',
				style = 'dropdown',
				values = { "Always", "Only if I've played both", "Never" },
				name = "Show Alliance / Horde timers:",
				desc = "Controls whether to show Alliance and Horde timers in the tooltip.",
				get = function(_) return db.profile.showFactions end,
				set = function(_, value) db.profile.showFactions = value end,
				order = 70,
			},
			reminderOptions = {
				type = "header",
				name = "Reminders",
				order = 200,
			},
			remindToggle = {
				type = 'toggle',
				name = 'Enable Reminders',
				desc = "",
				width = "full",
				get = function(_) return db.global.remindersEnabled end,
				set = function(_, value) db.global.remindersEnabled = value end,
				order = 210,
			},
			remindTimer = {
				type = 'select',
				style = 'dropdown',
				values = { "Time played this session", "Total played today", "This character today" },
				name = "Remind me when:",
				desc = "",
				get = function(_) return db.global.reminderTimer end,
				set = function(_, value) db.global.reminderTimer = value end,
				order = 220,
				disabled = function(_) return not db.global.remindersEnabled end,
			},
			remindAfter = {
				type = 'range',
				name = "exceeds (minutes):",
				desc = "Remind me after I've played for this length of time in minutes.",
				min = 15,
				max = 240,
				step = 15,
				get = function(_) return db.global.remindMinutes end,
				set = function(_, value) db.global.remindMinutes = value end,
				order = 230,
				disabled = function(_) return not db.global.remindersEnabled end,
			},
			snoozeToggle = {
				type = 'toggle',
				name = 'Remind me again',
				desc = "",
				get = function(_) return db.global.snoozeEnabled end,
				set = function(_, value) db.global.snoozeEnabled = value end,
				order = 240,
				disabled = function(_) return not db.global.remindersEnabled end,
			},
			snoozeAfter = {
				type = 'range',
				name = "every (minutes):",
				desc = "Snooze and remind me again after I've played for this length of time in minutes.",
				min = 15,
				max = 240,
				step = 15,
				get = function(_) return db.global.snoozeMinutes end,
				set = function(_, value) db.global.snoozeMinutes = value end,
				order = 250,
				disabled = function(_) 
					return not (db.global.snoozeEnabled and db.global.remindersEnabled)
				end,
			},
			reminderType = {
				type = 'select',
				style = 'radio',
				values = { "Whisper", "Raid Warning" },
				name = "Send my reminders using:",
				desc = "The reminder will be delivered via the specified channel.",
				get = function(_) return db.global.reminderType end,
				set = function(_, value) db.global.reminderType = value end,
				order = 260,
				disabled = function(_) return not db.global.remindersEnabled end,
			},
			reminderText = {
				type = 'input',
				name = "Reminder text:",
				width = 'full',
				multiline = true,
				get = function(_) return db.global.reminderText end,
				set = function(_, value) db.global.reminderText = value end,
				order = 270,
				disabled = function(_) return not db.global.remindersEnabled end,
			},
			reminderDescription = {
				type = 'description',
				name = "Use the string '%s' (without the quotes) to display the timer duration.",
				order = 280,
			},
		},
	})
	
	playTimerOptionsMenu = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Broker_PlayTimer", "PlayTimer", "Broker")


	sessionTimer = 0
	timerDisplayed = db.profile.defaultTimer
	startOfNewDay = tomorrow()
	
	if db.global.lastTimerUpdate < yesterday() then
		resetDayTimers()
	end
	createMyTimers()
	
	self:RegisterEvent("TIME_PLAYED_MSG")
	RequestTimePlayed();
	
	self:UpdateTimers()
	self:ScheduleRepeatingTimer("UpdateTimers", 60)			-- call UpdateTimers() once a minute
end


function tomorrow()
	local d = date("*t")
	
	if d.hour >= db.profile.firstHourOfNewDay then
		d.day = d.day + 1
	end
	
	d.hour, d.min, d.sec = db.profile.firstHourOfNewDay, 0, 0	
	return time(d)
end


function yesterday()
	local d = date("*t")
	
	if d.hour < db.profile.firstHourOfNewDay then
		d.day = d.day - 1
	end
	
	d.hour, d.min, d.sec = db.profile.firstHourOfNewDay, 0, 0	
	return time(d)
end


function resetDayTimers()
	db.global.todayTimer = 0
	db.global.realmTable = {}
end


function createMyTimers()
	if not thisRealm() then
		db.global.realmTable[realm] = { realmTimer = 0, playerTable = {} }
	end
	
	if not thisPlayer() then
		thisRealm().playerTable[player] = { playerTimer = 0, playerFaction = UnitFactionGroup("player") }
	end
end


function thisRealm()
	return db.global.realmTable[realm]
end


function thisPlayer()
	if thisRealm() then
		return thisRealm().playerTable[player]
	end
end


local function colorizeText(theText, theColor)
	if db.profile.showColorText and theColor then
		local r, g, b = theColor.r*255, theColor.g*255, theColor.b*255
		return string.format("|cff%02x%02x%02x%s|r", r, g, b, theText)
	else
		return theText
	end
end


function formatTime(theTime, theFormat)
	return formatTimeWithColor(theTime, theFormat, WHITE)
end


function formatTimeWithColor(theTime, theFormat, theColor)
	if not theTime then theTime = 0 end
	if not theFormat then theFormat = db.profile.timeFormat end
	local days, hours, minutes, dayString, hourString, minString, formatString
	local result = ""
	
	days = floor(theTime/1440)
	hours = floor((theTime % 1440)/60)
	minutes = theTime % 60
	
	if theFormat == SHORT_TIME_FORMAT then
		-- short time format is a special case because it displays "1h 0m" instead of just "1h"
		dayString, hourString, minString = "d", "h", "m"

		if days > 0 then
			formatString = colorizeText("%d", theColor)..dayString.." "
			result = (formatString):format(days)
		end
		
		if days > 0 or hours > 0 then
			formatString = colorizeText("%d", theColor)..hourString.." "
			result = result..(formatString):format(hours)
		end
		
		formatString = colorizeText("%d", theColor)..minString
		result = result..(formatString):format(minutes)
	else
		if theFormat == MED_TIME_FORMAT then
			dayString	= { [SINGULAR]="day",		[PLURAL]="days" }
			hourString	= { [SINGULAR]="hr",		[PLURAL]="hrs" }
			minString	= { [SINGULAR]="min",		[PLURAL]="mins" }
		else		-- LONG_TIME_FORMAT
			dayString	= { [SINGULAR]="day",		[PLURAL]="days" }
			hourString	= { [SINGULAR]="hour",		[PLURAL]="hours" }
			minString	= { [SINGULAR]="minute",	[PLURAL]="minutes" }		
		end
		
		if days > 0 then
			formatString = colorizeText("%d", theColor).." "..dayString[days==1]
			result = (formatString):format(days)
			if hours > 0 or minutes > 0 then result = result..", " end
		end
		
		if hours > 0 then
			formatString = colorizeText("%d", theColor).." "..hourString[hours==1]
			result = result..(formatString):format(hours)
			if minutes > 0 then result = result..", " end
		end
		
		if minutes > 0 then
			formatString = colorizeText("%d", theColor).." "..minString[minutes==1]
			result = result..(formatString):format(minutes)
		end		
	end	
	
	return result
end


function selectTimer(theTimer)
	local timerOptions = {
		sessionTimer, 
		db.global.todayTimer, 
		thisPlayer().playerTimer
	}
	return timerOptions[theTimer]
end


function playTimer:UpdateTimers()
	sessionTimer = sessionTimer + 1

	if time() >= startOfNewDay then
		startOfNewDay = tomorrow()
		resetDayTimers()
		createMyTimers()
	else
		db.global.todayTimer = db.global.todayTimer + 1
		thisPlayer().playerTimer = thisPlayer().playerTimer + 1
	end
	db.global.lastTimerUpdate = time()
	UpdateStatusBarText()
	
	if showingTooltip then
		GameTooltip:ClearLines()
		dataobj.OnTooltipShow(GameTooltip)
		GameTooltip:Show()
	end
	
	if db.global.remindersEnabled then
		local delta = selectTimer(db.global.reminderTimer) - db.global.remindMinutes
		if delta == 0 or (db.global.snoozeEnabled and delta % db.global.snoozeMinutes == 0) then
			local timeString	= formatTimeWithColor(selectTimer(db.global.reminderTimer), LONG_TIME_FORMAT, nil)
			local reminderText	= (db.global.reminderText):format(timeString)

			if db.global.reminderType == RAID_WARNING then
				RaidNotice_AddMessage(RaidWarningFrame, reminderText, ChatTypeInfo["RAID_WARNING"]);
				PlaySound("RaidWarning");
			elseif db.global.reminderType == WHISPER then
				DEFAULT_CHAT_FRAME:AddMessage("[PlayTimer addon] whispers: "..reminderText, PURPLE.r, PURPLE.g, PURPLE.b)
				PlaySound("TellMessage");
			end
		end
	end
	
end


function playTimer:TIME_PLAYED_MSG(msg, total, level)
	blizzTotalPlayed, blizzLevelPlayed = total/60, level/60
	self:UnregisterEvent("TIME_PLAYED_MSG")
end


function UpdateStatusBarText()
	local timerName = ""
	if db.profile.showTimerName then
		timerName = timerNames[timerDisplayed]..": "
	end
	dataobj.text = timerName..formatTime(selectTimer(timerDisplayed))
end


function dataobj:OnTooltipShow()
	self:AddLine(colorizeText("PlayTimer", ORANGE))
	self:AddLine(" ")
	
	-- Player
	self:AddLine(colorizeText(player, WHITE))
	self:AddDoubleLine("  ".."This Session", formatTime(sessionTimer))
	self:AddDoubleLine("  ".."Today", formatTime(thisPlayer().playerTimer))
	self:AddDoubleLine("  ".."Level".." "..UnitLevel(player), formatTime(blizzLevelPlayed + sessionTimer))
	self:AddDoubleLine("  ".."Lifetime", formatTime(blizzTotalPlayed + sessionTimer))
	self:AddLine(" ")

	-- Today
	self:AddLine(colorizeText("Today", WHITE))
	
	local numberOfRealms, factionTable, realmIndent = 0, {}, ""
	for realmName, theRealm in pairs(db.global.realmTable) do
		numberOfRealms = numberOfRealms + 1
		theRealm.realmTimer = 0
		for playerName, thePlayer in pairs(theRealm.playerTable) do
			theRealm.realmTimer = theRealm.realmTimer + thePlayer.playerTimer
			if not factionTable[thePlayer.playerFaction] then
				factionTable[thePlayer.playerFaction] = thePlayer.playerTimer
			else
				factionTable[thePlayer.playerFaction] = factionTable[thePlayer.playerFaction] + thePlayer.playerTimer
			end
		end
	end
	
	if numberOfRealms > 1 then
		realmIndent = "  "
	end
	
	for realmName, theRealm in pairs(db.global.realmTable) do
		if numberOfRealms > 1 then
			self:AddDoubleLine("  "..colorizeText(realmName, LTBLUE), formatTime(theRealm.realmTimer))
		end
		for playerName, thePlayer in pairs(theRealm.playerTable) do
			self:AddDoubleLine("  "..realmIndent..colorizeText(playerName, factionColor[thePlayer.playerFaction]), formatTime(thePlayer.playerTimer))
		end
	end

	
	if db.profile.showFactions == ALWAYS or (db.profile.showFactions == IF_BOTH_PLAYED and factionTable["Alliance"] and factionTable["Horde"]) then
		self:AddLine(" ")
		self:AddDoubleLine(colorizeText("  ".."Alliance", factionColor["Alliance"]), formatTime(factionTable["Alliance"]))
		self:AddDoubleLine(colorizeText("  ".."Horde",    factionColor["Horde"   ]), formatTime(factionTable["Horde"   ]))	
	end
	
	self:AddLine(" ")
	self:AddDoubleLine("  ".."Total Played Today", formatTime(db.global.todayTimer))
	
	
end


function dataobj:OnEnter()
	showingTooltip = true
	GameTooltip:SetOwner(self, "ANCHOR_NONE")
	GameTooltip:SetPoint("TOPLEFT", self, "BOTTOMLEFT")
	GameTooltip:ClearLines()
	dataobj.OnTooltipShow(GameTooltip)
	GameTooltip:Show()
end


function dataobj:OnLeave()
	showingTooltip = false
	GameTooltip:Hide()
end


function dataobj:OnClick(button)
	if button == "RightButton" then
		InterfaceOptionsFrame_OpenToCategory(playTimerOptionsMenu)
	elseif IsShiftKeyDown() then
		-- cycle between three timeFormat options
		db.profile.timeFormat = (db.profile.timeFormat % 3) + 1
	elseif IsControlKeyDown() then
		-- toggle display of timer name in status bar
		db.profile.showTimerName = not db.profile.showTimerName
	else
		-- cycle which timer is displayed in the status bar
		timerDisplayed = (timerDisplayed % 3) + 1
	end
	UpdateStatusBarText()
end